#!/usr/bin/env python3
#
# This file is part of Script of Scripts (sos), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/vatlab/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
##
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import subprocess
import tempfile
import copy
import urllib
import urllib.request
import urllib.error
import urllib.parse
import shutil
import zipfile
import shlex
import gzip
import tarfile
import fasteners
import uuid

from functools import wraps

from collections.abc import Sequence
import multiprocessing as mp
from tqdm import tqdm as ProgressBar
from .utils import env, transcribe, StopInputGroup, TerminateExecution, short_repr, get_traceback
from .sos_eval import Undetermined, interpolate
from .target import FileTarget, fileMD5, executable, UnknownTarget, BaseTarget


__all__ = ['SoS_Action', 'script', 'sos_run',
    'fail_if', 'warn_if', 'stop_if',
    'download',
    'run', 'perl', 'ruby', 
    'report', 'pandoc'
    ]

from .sos_syntax import SOS_RUNTIME_OPTIONS, SOS_ACTION_OPTIONS
from .sos_script import SoS_Script

def get_actions():
    # get the name of all actions, which are identified by an attribute
    # run_mode of the function
    return [k for k, v in globals().items() if hasattr(v, 'run_mode')]

#
# A decoration function that allows SoS to replace all SoS actions
# with a null action.
#
def SoS_Action(run_mode=('run', 'interactive'), acceptable_args=('*',)):
    run_mode = [run_mode] if isinstance(run_mode, str) else run_mode
    def runtime_decorator(func):
        @wraps(func)
        def action_wrapper(*args, **kwargs):
            # if docker_image in args, a large number of docker-specific
            # args would be allowed.
            if '*' not in acceptable_args and 'docker_image' not in kwargs:
                for key in kwargs.keys():
                    if key not in acceptable_args and key not in SOS_ACTION_OPTIONS:
                        raise ValueError('Unrecognized option "{}" for action {}'.format(key, func))
            # docker files will be downloaded in run or prepare mode
            if 'docker_file' in kwargs and env.config['run_mode'] in ['run', 'interactive']:
                from .docker.client import SoS_DockerClient
                docker = SoS_DockerClient()
                docker.load_image(kwargs['docker_file'])
            # handle image
            if 'docker_image' in kwargs:
                from .docker.client import SoS_DockerClient
                docker = SoS_DockerClient()
                docker.pull(kwargs['docker_image'])
            if env.config['run_mode'] not in run_mode:
                # return dynamic expression when not in run mode, that is to say
                # the script logic cannot rely on the result of the action
                return Undetermined(func.__name__)
            if env.config['run_mode'] == 'interactive':
                for k,v in kwargs.items():
                    if k in SOS_RUNTIME_OPTIONS and k not in SOS_ACTION_OPTIONS:
                        env.logger.warning('Passing runtime option "{0}" to action is deprecated. Please use "task: {0}={1}" before action instead.'.format(k, v))
            if 'active' in kwargs:
                if isinstance(kwargs['active'], int):
                    if kwargs['active'] >= 0 and env.sos_dict['_index'] != kwargs['active']:
                        return None
                    if kwargs['active'] < 0 and env.sos_dict['_index'] != kwargs['active'] + env.sos_dict['__num_groups__']:
                        return None
                elif isinstance(kwargs['active'], Sequence):
                    allowed_index = list([x if x >= 0 else env.sos_dict['__num_groups__'] + x for x in kwargs['active']])
                    if env.sos_dict['_index'] not in allowed_index:
                        return None
                elif isinstance(kwargs['active'], slice):
                    allowed_index = list(range(env.sos_dict['__num_groups__']))[kwargs['active']]
                    if env.sos_dict['_index'] not in allowed_index:
                        return None
                else:
                    raise RuntimeError('Unacceptable value for option active: {}'.format(kwargs['active']))
            # if there are parameters input and output, the action is subject to signature verification
            sig = None
            if 'input' in kwargs and 'output' in kwargs:
                if args and isinstance(args[0], str):
                    script = args[0]
                elif 'script' in kwargs:
                    script = kwargs['script']
                else:
                    script = ''
                # 
                from .target import RuntimeInfo
                sig = RuntimeInfo(func.__name__, script, [kwargs['input']] if isinstance(kwargs['input'], str) else kwargs['input'],
                        [kwargs['output']] if isinstance(kwargs['output'], str) else kwargs['output'], [], kwargs)
                sig.lock()
                if env.config['sig_mode'] == 'default':
                    matched = sig.validate()
                    if isinstance(matched, dict):
                        env.logger.info('Action ``{}`` is ``ignored`` due to saved signature'.format(func.__name__))
                        return None
                    else:
                        env.logger.debug('Signature mismatch: {}'.format(matched))
                elif env.config['sig_mode'] == 'assert':
                    matched = sig.validate()
                    if isinstance(matched, str):
                        raise RuntimeError('Signature mismatch: {}'.format(matched))
                    else:
                        env.logger.info('Action ``{}`` is ``ignored`` with matching signature'.format(func.__name__))
                        return None
                elif env.config['sig_mode'] == 'build':
                    # build signature require existence of files
                    if sig.write(rebuild=True):
                        env.logger.info('Action ``{}`` is ``ignored`` with signature constructed'.format(func.__name__))
                        return None
            if 'workdir' in kwargs:
                if not kwargs['workdir'] or not isinstance(kwargs['workdir'], str):
                    raise RuntimeError('workdir option should be a path, {} provided'.format(kwargs['workdir']))
                if not os.path.isdir(os.path.expanduser(kwargs['workdir'])):
                    os.makedirs(os.path.expanduser(kwargs['workdir']))
                try:
                    olddir = os.getcwd()
                    os.chdir(os.path.expanduser(kwargs['workdir']))
                    try:
                        res = func(*args, **kwargs)
                    except Exception as e:
                        if 'allow_error' in kwargs and kwargs['allow_error']:
                            env.logger.warning(e)
                            res = None
                        else:
                            raise
                finally:
                    os.chdir(olddir)
            else:
                try:
                    res = func(*args, **kwargs)
                except Exception as e:
                    if 'allow_error' in kwargs and kwargs['allow_error']:
                        env.logger.warning(e)
                        res = None
                    else:
                        raise
            if 'output' in kwargs:
                ofiles = [kwargs['output']] if isinstance(kwargs['output'], str) else kwargs['output']
                for ofile in ofiles:
                    if isinstance(ofile, str):
                        if not FileTarget(ofile).exists('any'):
                            raise RuntimeError('Output target {} does not exist after completion of action {}'.format(ofile, func.__name__))
                    elif isinstance(ofile, BaseTarget):
                        if not ofile.exists('any'):
                            raise RuntimeError('Output target {} does not exist after completion of action {}'.format(ofile, func.__name__))
                    else:
                        raise ValueError('Unrecognized output target {} with type {}'.format(ofile, ofile.__class__.__name__))
            if sig:
                sig.write()
                sig.release()
            return res
        action_wrapper.run_mode = run_mode
        return action_wrapper
    return runtime_decorator


class SoS_ExecuteScript:
    def __init__(self, script, interpreter, suffix, args=''):
        self.script = script
        self.interpreter = interpreter
        self.args = args
        if suffix:
            self.suffix = suffix
        elif sys.platform == 'win32':
            self.suffix = '.bat'
        else:
            self.suffix = '.sh'

    def run(self, **kwargs):
        #
        if 'input' in kwargs:
            if isinstance(kwargs['input'], str):
                ifiles = [kwargs['input']]
            elif isinstance(kwargs['input'], Sequence):
                ifiles = list(kwargs['input'])
            else:
                raise ValueError('Unacceptable value for paremter input: {}'.format(kwargs['input']))

            content = ''
            for ifile in ifiles:
                with open(ifile) as iscript:
                    content += iscript.read()
            self.script = content + self.script

        if 'docker_image' in kwargs:
            if env.config['run_mode'] == 'dryrun':
                print('In docker image {}\n{}:\n{}\n'.format(kwargs['docker_image'], self.interpreter, self.script))
                return None
            from .docker.client import SoS_DockerClient
            docker = SoS_DockerClient()
            docker.run(kwargs['docker_image'], self.script, self.interpreter, self.args, self.suffix,
                **kwargs)
        else:
            if isinstance(self.interpreter, str):
                if self.interpreter and not shutil.which(shlex.split(self.interpreter)[0]):
                    raise RuntimeError('Failed to locate interpreter {}'.format(self.interpreter))
            elif isinstance(self.interpreter, Sequence):
                found = False
                for ip in self.interpreter:
                    if shutil.which(shlex.split(ip)[0]):
                        self.interpreter = ip
                        found = True
                        break
                if not found:
                    raise RuntimeError('Failed to locate any of the interpreters {}'
                        .format(', '.join(self.interpreter)))
            else:
                raise RuntimeError('Unacceptable interpreter {}'.format(self.interpreter))

            transcribe(self.script, action=self.interpreter)
            debug_script_file = os.path.join(env.exec_dir, '.sos', '{}_{}_{}{}'.format(env.sos_dict['step_name'],
                env.sos_dict['_index'], str(uuid.uuid4())[:8], self.suffix))
            #env.logger.debug('Script for step {} is saved to {}'.format(env.sos_dict['step_name'], debug_script_file))
            with open(debug_script_file, 'w') as sfile:
                sfile.write(self.script)
            env.logger.trace(self.script)

            try:
                p = None
                script_file = tempfile.NamedTemporaryFile(mode='w+t', suffix=self.suffix, delete=False).name
                with open(script_file, 'w') as sfile:
                    sfile.write(self.script)
                if self.interpreter:
                    # if there is an interpreter and with args
                    if not self.args:
                        self.args = '${filename!q}'
                else:
                    if sys.platform == 'win32':
                        # in the case there is no interpreter, we put the script
                        # at first (this is the case for windows)
                        #
                        # and we donot add default args.
                        self.interpreter = '${filename!q}'
                    else:
                        # if there is a shebang line, we ...
                        if self.script.startswith('#!'):
                            # make the script executable
                            os.chmod(script_file, 0o775)
                            self.interpreter = '${filename!q}'
                        else:
                            self.interpreter = '/bin/bash'
                            if not self.args:
                                self.args = '-ev ${filename!q}'
                #
                if env.config['run_mode'] == 'dryrun':
                    print('{}:\n{}\n'.format(self.interpreter, self.script))
                    return None
                cmd = interpolate('{} {}'.format(self.interpreter, self.args), '${ }', {'filename': script_file, 'script': self.script})
                #
                if env.config['run_mode'] == 'interactive':
                    # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
                    from .utils import pexpect_run
                    ret = pexpect_run(cmd)
                elif '__std_out__' in env.sos_dict and '__std_err__' in env.sos_dict:
                    if env.verbosity >= 1:
                        with open(env.sos_dict['__std_out__'], 'ab') as so, open(env.sos_dict['__std_err__'], 'ab') as se:
                            p = subprocess.Popen(cmd, shell=True, stderr=se, stdout=so)
                    else:
                        p = subprocess.Popen(cmd, shell=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
                    ret = p.wait()
                    if ret != 0:
                        task_id = os.path.basename(env.sos_dict['__std_err__']).split('.')[0]
                        raise RuntimeError('Use "sos status {} -v4" for details.'.format(task_id))
                else:
                    p = subprocess.Popen(cmd, shell=True,
                                         stderr=None if env.verbosity > 0 else subprocess.DEVNULL,
                                         stdout=None if env.verbosity > 1 else subprocess.DEVNULL)
                    ret = p.wait()
                if ret != 0:
                    with open(debug_script_file, 'w') as sfile:
                        sfile.write(self.script)
                    if self.interpreter == '/bin/bash' and self.args == '-ev ${filename!q}':
                        debug_args = '${filename!q}'
                    else:
                        debug_args = self.args
                    cmd = interpolate('{} {}'.format(self.interpreter, debug_args), '${ }', {'filename': debug_script_file, 'script': self.script})
                    raise RuntimeError('Failed to execute commmand ``{}`` (ret={}, workdir={})'.format(cmd, ret, os.getcwd()))
            except RuntimeError:
                raise
            except Exception as e:
                env.logger.error('Failed to execute script: {}'.format(e))
                raise
            finally:
                os.remove(script_file)


@SoS_Action(run_mode=['run', 'interactive'])
def sos_run(workflow=None, targets=None, shared=None, args=None, **kwargs):
    '''Execute a workflow from specified source, input, and output
    By default the workflow is defined in the existing SoS script, but
    extra sos files can be specified from paramter source. The workflow
    will be execute in the current step namespace with _input as workflow
    input. '''
    if '__std_out__' in env.sos_dict and '__std_err__' in env.sos_dict:
        raise RuntimeError('Executing nested workflow (action sos_run) in tasks is not supported.')

    script = SoS_Script(env.sos_dict['__step_context__'].content, env.sos_dict['__step_context__'].filename)
    wf = script.workflow(workflow, use_default=not targets)
    # if wf contains the current step or one of the previous one, this constitute
    # recusive nested workflow and should not be allowed
    if env.sos_dict['step_name'] in ['{}_{}'.format(x.name, x.index) for x in wf.sections]:
        raise RuntimeError('Nested workflow {} contains the current step {}'.format(workflow, env.sos_dict['step_name']))
    # args can be specified both as a dictionary or keyword arguments
    if args is None:
        args = kwargs
    else:
        args.update(kwargs)
    args['__args__'] = env.sos_dict['__args__']
    if shared is None:
        shared = []
    elif isinstance(shared, str):
        shared = [shared]

    # for nested workflow, _input would becomes the input of workflow.
    env.sos_dict.set('__step_output__', copy.deepcopy(env.sos_dict.get('_input', None)))
    shared.append('__step_output__')
    try:
        my_name = env.sos_dict['step_name']
        args_output = ', '.join('{}={}'.format(x, short_repr(y)) for x,y in args.items() if not x.startswith('__'))
        env.logger.debug('Executing workflow ``{}`` with input ``{}`` and {}'
            .format(workflow, short_repr(env.sos_dict.get('_input', None), True),
            'no args' if not args_output else args_output))

        if not hasattr(env, '__pipe__'):
            # if env has no __pipe__, this means the nested workflow is executed from
            # within a task, and we can just use an executor to execute it.
            #         # tell the master process to receive a workflow
            shared = {x: (env.sos_dict[x] if x in env.sos_dict else None) for x in shared}
            if env.config['run_mode'] == 'run':
                from .sos_executor import Base_Executor
                executor = Base_Executor(wf, args=args, shared=shared, config=env.config)
                if shared:
                    q = mp.Pipe()
                else:
                    q = None
                p = mp.Process(target=executor.run, kwargs={'targets': targets, 'parent_pipe': q[1], 'my_workflow_id': None})
                p.start()
                if shared:
                    res = q[0].recv()
                    if res is None:
                        # worker is killed
                        sys.exit(0)
                    elif isinstance(res, Exception):
                        raise res
                    env.sos_dict.quick_update(res['shared'])
                else:
                    res = None
                p.join()
            else:
                from sos_notebook.sos_executor import Interactive_Executor
                executor = Interactive_Executor(wf, args=args, shared=shared, config=env.config)
                res = executor.run(targets=targets)
            return res

        else:
            # tell the master process to receive a workflow
            env.__pipe__.send('workflow {}'.format(uuid.uuid4()))
            # really send the workflow
            shared = {x: (env.sos_dict[x] if x in env.sos_dict else None) for x in shared}
            env.__pipe__.send((wf, targets, args, shared, env.config))
            res = env.__pipe__.recv()
            if res is None:
                sys.exit(0)
            elif isinstance(res, Exception):
                raise res
            else:
                env.sos_dict.quick_update(res['shared'])
                return res
    finally:
        # restore step_name in case the subworkflow re-defines it
        env.sos_dict.set('step_name', my_name)

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['script', 'interpreter', 'suffix', 'args'])
def script(script, interpreter, suffix='', args='', **kwargs):
    '''Execute specified script using specified interpreter. This action accepts common
    action arguments such as input, active, workdir, docker_image and args. In particular,
    content of one or more files specified by option input would be prepended before
    the specified script.'''
    if env.config['run_mode'] == 'dryrun':
        print('{}:\n{}\n'.format(interpreter, script))
        return None
    return SoS_ExecuteScript(script, interpreter, suffix, args).run(**kwargs)

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['expr', 'msg'])
def fail_if(expr, msg=''):
    '''Raise an exception with `msg` if condition `expr` is False'''
    if expr:
        raise TerminateExecution(msg if msg else 'error triggered by action fail_if')
    return 0

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['expr', 'msg'])
def warn_if(expr, msg=''):
    '''Yield an warning message `msg` if `expr` is False '''
    if expr:
        env.logger.warning(msg)
    return 0

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['expr', 'msg'])
def stop_if(expr, msg=''):
    '''Abort the execution of the current step or loop and yield
    an warning message `msg` if `expr` is False '''
    if expr:
        raise StopInputGroup(msg)
    return 0

#
# download file with progress bar
#
def downloadURL(URL, dest, decompress=False, index=None):
    dest = os.path.abspath(os.path.expanduser(dest))
    dest_dir, filename = os.path.split(dest)
    #
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    if not os.path.isdir(dest_dir):
        raise RuntimeError('Failed to create destination directory to download {}'.format(URL))
    #
    message = filename
    if len(message) > 30:
        message = message[:10] + '...' + message[-16:]
    #
    dest_tmp = dest + '.tmp_{}'.format(os.getpid())
    term_width = shutil.get_terminal_size((80, 20)).columns
    try:
        env.logger.debug('Download {} to {}'.format(URL, dest))
        sig = FileTarget(dest)
        if os.path.isfile(dest):
            prog = ProgressBar(desc=message + ': \033[32m validating\033[0m', disable=env.verbosity <= 1,
                position=index, leave=True, bar_format='{desc}', total=10000000)
            if env.config['sig_mode'] == 'build':
                if decompress:
                    prog.set_description(message + ': \033[32m scanning decompressed files\033[0m')
                    prog.update()
                    if zipfile.is_zipfile(dest):
                        zfile = zipfile.ZipFile(dest)
                        names = zfile.namelist()
                        for name in names:
                            # only python3.6 has the is_dir function for ZipInfo
                            if name.endswith('/'):
                                continue
                            dest_file = os.path.join(dest_dir, name)
                            if not os.path.isfile(dest_file):
                                env.logger.warning('Missing decompressed file {}'.format(dest_file))
                            else:
                                sig.add(dest_file)
                    elif tarfile.is_tarfile(dest):
                        with tarfile.open(dest, 'r:*') as tar:
                            # only extract files
                            file_count = 0
                            for tarinfo in tar:
                                if tarinfo.isfile():
                                    dest_file = os.path.join(dest_dir, tarinfo.name)
                                    if not os.path.isfile(dest_file):
                                        env.logger.warning('Missing decompressed file {}'.format(dest_file))
                                    else:
                                        sig.add(dest_file)
                                        file_count += 1
                                # sometimes the file is very large but we do not need to decompress all
                                # and track all files.
                                if file_count > 10:
                                    break
                    elif dest.endswith('.gz'):
                        decomp = dest[:-3]
                        if not os.path.isfile(decomp):
                            env.logger.warning('Missing decompressed file {}'.format(decomp))
                        sig.add(decomp)
                prog.set_description(message + ': \033[32m writing signature\033[0m')
                prog.update()
                sig.write_sig()
                prog.set_description(message + ': \033[32m signature calculated\033[0m')
                prog.update()
                prog.close()
                return True
            elif env.config['sig_mode'] == 'ignore':
                prog.set_description(message + ': \033[32m use existing\033[0m')
                prog.update()
                prog.close()
                return True
            elif env.config['sig_mode'] == 'default':
                prog.update()
                if sig.validate():
                    prog.set_description(message + ': \033[32m Validated\033[0m')
                    prog.update()
                    prog.close()
                    return True
                else:
                    prog.set_description(message + ':\033[91m Signature mismatch\033[0m')
                    prog.update()
        else:
            prog = ProgressBar(desc=message, disable=env.verbosity <= 1, position=index,
                leave=True, bar_format='{desc}', total=10000000)

        #
        # Stop using pycurl because of libcurl version compatibility problems
        # that happen so often and difficult to fix. Error message looks like
        #
        # Reason: Incompatible library version: pycurl.cpython-35m-darwin.so
        # requires version 9.0.0 or later, but libcurl.4.dylib provides version 7.0.0
        #
        #with open(dest_tmp, 'wb') as f:
        #    c = pycurl.Curl()
        #    c.setopt(pycurl.URL, str(URL))
        #    c.setopt(pycurl.WRITEFUNCTION, f.write)
        #    c.setopt(pycurl.SSL_VERIFYPEER, False)
        #    c.setopt(pycurl.NOPROGRESS, False)
        #    c.setopt(pycurl.PROGRESSFUNCTION, prog.curlUpdate)
        #    c.perform()
        #if c.getinfo(pycurl.HTTP_CODE) == 404:
        #    prog.set_description(message + ':\033[91m 404 Error {}\033[0m'.format(' '*(term_width - len(message) - 12)))
        #    try:
        #        os.remove(dest_tmp)
        #    except OSError:
        #        pass
        #    return False
        with open(dest_tmp, 'wb') as f:
            try:
                u = urllib.request.urlopen(str(URL))
                try:
                    file_size = int(u.getheader("Content-Length"))
                    prog = ProgressBar(total=file_size, desc=message, position=index, leave=False)
                except Exception:
                    file_size = None
                file_size_dl = 0
                block_sz = 8192
                while True:
                    buffer = u.read(block_sz)
                    if not buffer:
                        break
                    file_size_dl += len(buffer)
                    f.write(buffer)
                    prog.update(len(buffer))
            except urllib.error.HTTPError as e:
                prog.set_description(message + ':\033[91m {} Error\033[0m'.format(e.code))
                prog.update()
                prog.close()
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
            except Exception as e:
                prog.set_description(message + ':\033[91m {}\033[0m'.format(e))
                prog.update()
                prog.close()
                try:
                    os.remove(dest_tmp)
                except OSError:
                    pass
                return False
        #
        os.rename(dest_tmp, dest)
        decompressed = 0
        if decompress:
            if zipfile.is_zipfile(dest):
                prog.set_description(message + ':\033[91m Decompressing\033[0m')
                prog.update()
                prog.close()
                zfile = zipfile.ZipFile(dest)
                zfile.extractall(dest_dir)
                names = zfile.namelist()
                for name in names:
                    if os.path.isdir(os.path.join(dest_dir, name)):
                        continue
                    elif not os.path.isfile(os.path.join(dest_dir, name)):
                        return False
                    else:
                        sig.add(os.path.join(dest_dir, name))
                        decompressed += 1
            elif tarfile.is_tarfile(dest):
                prog.set_description(message + ':\033[91m Decompressing\033[0m')
                prog.update()
                prog.close()
                with tarfile.open(dest, 'r:*') as tar:
                    tar.extractall(dest_dir)
                    # only extract files
                    files = [x.name for x in tar.getmembers() if x.isfile()]
                    for name in files:
                        if not os.path.isfile(os.path.join(dest_dir, name)):
                            return False
                        else:
                            sig.add(os.path.join(dest_dir, name))
                            decompressed += 1
            elif dest.endswith('.gz'):
                prog.set_description(message + ':\033[91m Decompressing\033[0m')
                prog.update()
                prog.close()
                decomp = dest[:-3]
                with gzip.open(dest, 'rb') as fin, open(decomp, 'wb') as fout:
                    buffer = fin.read(100000)
                    while buffer:
                        fout.write(buffer)
                        buffer = fin.read(100000)
                sig.add(decomp)
                decompressed += 1
        decompress_msg = '' if not decompressed else ' ({} file{} decompressed)'.format(
            decompressed, '' if decompressed <= 1 else 's')
        prog.set_description(message + ':\033[32m downloaded{} {}\033[0m'.format(decompress_msg,
            ' '*(term_width - len(message) - 13 - len(decompress_msg))))
        prog.update()
        prog.close()
        # if a md5 file exists
        # if downloaded files contains .md5 signature, use them to validate
        # downloaded files.
        if os.path.isfile(dest + '.md5'):
            prog.set_description(message + ':\033[91m Verifying md5 signature\033[0m')
            prog.update()
            prog.close()
            with open(dest + '.md5') as md5:
                rec_md5 = md5.readline().split()[0].strip()
                obs_md5 = fileMD5(dest, partial=False)
                if rec_md5 != obs_md5:
                    prog.set_description(message + ':\033[91m MD5 signature mismatch\033[0m')
                    prog.update()
                    prog.close()
                    env.logger.warning('md5 signature mismatch for downloaded file {} (recorded {}, observed {})'
                        .format(filename[:-4], rec_md5, obs_md5))
            prog.set_description(message + ':\033[91m MD5 signature verified\033[0m')
            prog.update()
            prog.close()
    except Exception as e:
        if env.verbosity > 2:
             sys.stderr.write(get_traceback())
        env.logger.error('Failed to download: {}'.format(e))
        return False
    finally:
        # if there is something wrong still remove temporary file
        if os.path.isfile(dest_tmp):
            os.remove(dest_tmp)
    sig.write_sig()
    return os.path.isfile(dest)


@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['URLs', 'dest_dir', 'dest_file', 'decompress'])
def download(URLs, dest_dir='.', dest_file=None, decompress=False):
    '''Download files from specified URL, which should be space, tab or
    newline separated URLs. The files will be downloaded to specified
    destination. If `filename.md5` files are downloaded, they are used to
    validate downloaded `filename`. Unless otherwise specified, compressed
    files are decompressed.
    '''
    if env.config['run_mode'] == 'dryrun':
        print('download\n{}\n'.format(URLs))
        return None
    if isinstance(URLs, str):
        urls = [x.strip() for x in URLs.split() if x.strip()]
    else:
        urls = list(URLs)
    #
    if dest_file is not None and len(urls) != 1:
        raise RuntimeError('Only one URL is allowed if a destination file is specified.')
    #
    if dest_file is None:
        filenames = []
        for idx, url in enumerate(urls):
            token = urllib.parse.urlparse(url)
            # if no scheme or netloc, the URL is not acceptable
            if not all([getattr(token, qualifying_attr) for qualifying_attr in  ('scheme', 'netloc')]):
                filenames.append(None)
                continue
            filename = os.path.split(token.path)[-1]
            if not filename:
                filenames.append(None)
                continue
            filenames.append(os.path.join(dest_dir, filename))
    else:
        filenames = [dest_file]
    #
    succ = [False for x in urls]
    if len(succ) > 1:
        # first scroll several lines to reserve place for progress bar
        with mp.Pool(processes = env.sos_dict['CONFIG'].get('sos_download_processes', 5)) as pool:
            for idx, (url, filename) in enumerate(zip(urls, filenames)):
                if not filename:
                    continue
                succ[idx] = pool.apply_async(downloadURL, (url, filename,
                    decompress, idx))
            succ = [x.get() if isinstance(x, mp.pool.AsyncResult) else x for x in succ]
    else:
        if dest_file is not None:
            succ[0] = downloadURL(urls[0], dest_file, decompress=decompress)
        else:
           if filenames[0]:
                succ[0] = downloadURL(urls[0], filenames[0], decompress=decompress)
    #
    #for su, url in zip(succ, urls):
    #    if not su:
    #        env.logger.warning('Failed to download {}'.format(url))
    if not all(succ):
        raise RuntimeError('Not all files have been downloaded')
    return 0

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['script', 'args'])
def run(script, args='', **kwargs):
    '''Execute specified script using bash. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, '', '', args).run(**kwargs)

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['script', 'args'])
def perl(script, args='', **kwargs):
    '''Execute specified script using perl. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, 'perl', '.pl', args).run(**kwargs)

@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['script', 'args'])
def ruby(script, args='', **kwargs):
    '''Execute specified script using ruby. This action accepts common action arguments such as
    input, active, workdir, docker_image and args. In particular, content of one or more files
    specified by option input would be prepended before the specified script.'''
    return SoS_ExecuteScript(script, 'ruby', '.rb', args).run(**kwargs)


def collect_input(script, input):
    # determine file extension
    if input is not None:
        if isinstance(input, str):
            ext = os.path.splitext(input)[-1]
        elif isinstance(input, Sequence) and len(input) > 0:
            ext = os.path.splitext(input[0])[-1]
        else:
            raise ValueError('Unknown input file for action pandoc')
    else:
        ext = '.md'

    input_file = tempfile.NamedTemporaryFile(mode='w+t', suffix=ext, delete=False).name
    with open(input_file, 'w') as tmp:
        if script is not None and script.strip():
            tmp.write(script.rstrip() + '\n\n')
        if isinstance(input, str):
            try:
                with open(input) as ifile:
                    tmp.write(ifile.read() + '\n\n')
            except  Exception as e:
                    raise ValueError('Failed to read input file {}: {}'.format(input, e))
        elif isinstance(input, Sequence):
            for ifile in input:
                try:
                    with open(ifile) as itmp:
                        tmp.write(itmp.read().rstrip() + '\n\n')
                except  Exception as e:
                        raise ValueError('Failed to read input file {}: {}'.format(ifile, e))
    return input_file


@SoS_Action(run_mode=['dryrun', 'run', 'interactive'], acceptable_args=['script'])
def report(script=None, input=None, output=None, **kwargs):
    '''Write script to an output file specified by `output`, which can be
    a filename to which the content of the script will be written,
    any object with a "write" attribute (e.g. a file handle) for which the "write"
    function will be called with the content. If output is unspecified, the content
    will be written to standard output or appended to a file specified with command
    line option `-r`. '''
    if env.config['run_mode'] == 'dryrun':
        print('report:\n{}'.format('' if script is None else script))
        if input is not None:
            for ifile in input:
                print('  from file: {}'.format(ifile))
        return
    file_handle = None
    if isinstance(output, str):
        if not output or output == '-':
            writer = sys.stdout.write
        elif output.startswith('>>'):
            file_handle = open(os.path.expanduser(output[2:]), 'a')
            writer = file_handle.write
        else:
            file_handle = open(os.path.expanduser(output), 'w')
            writer = file_handle.write
    elif hasattr(output, 'write'):
        writer = output.write
    elif output is None or output == '':
        writer = sys.stdout.write
    else:
        raise ValueError('Invalid output {}.'.format(output))

    # file lock to prevent race condition
    with fasteners.InterProcessLock(os.path.join(tempfile.gettempdir(), 'report_lock')):
        if isinstance(script, str) and script.strip():
            writer(script.rstrip() + '\n\n')
        if input is not None:
            if isinstance(input, str):
                env.logger.debug('Loading report from {}'.format(input))
                with open(input) as ifile:
                    writer(ifile.read().rstrip() + '\n\n')
            elif isinstance(input, Sequence):
                for ifile in input:
                    try:
                        env.logger.debug('Loading report from {}'.format(ifile))
                        with open(ifile) as itmp:
                            writer(itmp.read().rstrip() + '\n\n')
                    except Exception as e:
                        raise ValueError('Failed to read input file {}: {}'.format(ifile, e))
            else:
                raise ValueError('Unknown input file for action report')
    #
    if file_handle:
        file_handle.close()


@SoS_Action(run_mode=['run', 'interactive'], acceptable_args=['script', 'args'])
def pandoc(script=None, input=None, output=None, args='${input!q} --output ${output!q}', **kwargs):
    '''Convert input file to output using pandoc

    The input can be specified in three ways:

    1. instant script, which is assumed to be in md format

    pandoc:   output='report.html'
      script

    2. one or more input files. The format is determined by extension of input file

    pandoc(input, output='report.html')

    3. input file specified by command line option `-r` .
    pandoc(output='report.html')

    If no output is specified, it is assumed to be in html format
    and is written to standard output.

    You can specify more options such as "from" and "to" by customizing
    the args parameter of the action. The default value of args is
    `${input!q} --output ${output!q}'
    '''
#
#     # this is output format
#     pandoc [OPTIONS] [FILES]
# Input formats:  commonmark, docbook, docx, epub, haddock, html, json*, latex,
#                 markdown, markdown_github, markdown_mmd, markdown_phpextra,
#                 markdown_strict, mediawiki, native, odt, opml, org, rst, t2t,
#                 textile, twiki
#                 [ *only Pandoc's JSON version of native AST]
# Output formats: asciidoc, beamer, commonmark, context, docbook, docx, dokuwiki,
#                 dzslides, epub, epub3, fb2, haddock, html, html5, icml, json*,
#                 latex, man, markdown, markdown_github, markdown_mmd,
#                 markdown_phpextra, markdown_strict, mediawiki, native, odt,
#                 opendocument, opml, org, pdf**, plain, revealjs, rst, rtf, s5,
#                 slideous, slidy, tei, texinfo, textile
#                 [**for pdf output, use latex or beamer and -o FILENAME.pdf]
# Options:
#   -f FORMAT, -r FORMAT  --from=FORMAT, --read=FORMAT
#   -t FORMAT, -w FORMAT  --to=FORMAT, --write=FORMAT
#   -o FILENAME           --output=FILENAME
#                         --data-dir=DIRECTORY
#   -R                    --parse-raw
#   -S                    --smart
#
# IGNORED
#
    if not executable('pandoc').exists():
        raise UnknownTarget(executable('pandoc'))

    input_file = collect_input(script, input)

    write_to_stdout = False
    if output is None:
        write_to_stdout = True
        output_file = tempfile.NamedTemporaryFile(mode='w+t', suffix='.html', delete=False).name
    elif isinstance(output, str):
        output_file = os.path.expanduser(output)
    else:
        raise RuntimeError('A filename is expected, {} provided'.format(output))
    #
    ret = 1
    try:
        p = None
        cmd = interpolate('pandoc {}'.format(args), '${ }', {'input': input_file, 'output': output_file})
        env.logger.trace('Running command "{}"'.format(cmd))
        if env.config['run_mode'] == 'interactive':
            # need to catch output and send to python output, which will in trun be hijacked by SoS notebook
            from .utils import pexpect_run
            ret = pexpect_run(cmd)
        else:
            p = subprocess.Popen(cmd, shell=True)
            ret = p.wait()
    except Exception as e:
        env.logger.error(e)
    if ret != 0:
        temp_file = os.path.join('.sos', '{}_{}.md'.format('pandoc', os.getpid()))
        shutil.copyfile(input_file, temp_file)
        cmd = interpolate('pandoc {}'.format(args), '${ }', {'input': temp_file, 'output': output_file})
        raise RuntimeError('Failed to execute script. Please use command \n{}\nunder {} to test it.'
            .format(cmd, os.getcwd()))
    if write_to_stdout:
        with open(output_file) as out:
            sys.stdout.write(out.read())
    else:
        env.logger.info('Report saved to {}'.format(output))
    try:
        os.remove(input_file)
    except Exception:
        pass




#!/usr/bin/env python3
#
# This file is part of Script of Scripts (SoS), a workflow system
# for the execution of commands and scripts in different languages.
# Please visit https://github.com/vatlab/SOS for more information.
#
# Copyright (C) 2016 Bo Peng (bpeng@mdanderson.org)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
import hashlib
import shlex
import shutil
import fasteners
import pkg_resources

from collections.abc import Sequence

from .utils import env, Error, short_repr, stable_repr, save_var, load_var, isPrimitive
from .sos_eval import Undetermined

__all__ = ['dynamic', 'executable', 'env_variable', 'sos_variable']

class UnknownTarget(Error):
    def __init__(self, target):
        Error.__init__(self, 'Target unavailable: %s' % target)
        self.target = target

class RemovedTarget(Error):
    def __init__(self, target):
        Error.__init__(self, 'Target removed: %s' % target)
        self.target = target

class UnavailableLock(Error):
    """Raised when there are errors in prepare mode. Such errors are not raised
    immediately, but will be collected and raised at the end """

    def __init__(self, signature):
        Error.__init__(self, ('Failed to obtain a lock for input {} and output {}. It is likely ' +
            'that these files are protected by another SoS process or concurrant task ' +
            'that is generating the same set of files.').format(short_repr(signature[0]),
                short_repr(signature[1])))
        self.input = signature[0]
        self.output = signature[1]
        self.sig_file = signature[2]

#
# Runtime signature
#
def textMD5(text):
    '''Get md5 of a piece of text'''
    m = hashlib.md5()
    if isinstance(text, str):
        m.update(text.encode())
    else:
        m.update(text)
    return m.hexdigest()

def fileMD5(filename, partial=True):
    '''Calculate partial MD5, basically the first and last 8M
    of the file for large files. This should signicicantly reduce
    the time spent on the creation and comparison of file signature
    when dealing with large bioinformat ics datasets. '''
    filesize = os.path.getsize(filename)
    # calculate md5 for specified file
    md5 = hashlib.md5()
    block_size = 2**20  # buffer of 1M
    try:
        # 2**24 = 16M
        if (not partial) or filesize < 2**24:
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    if not data:
                        break
                    md5.update(data)
        else:
            count = 16
            # otherwise, use the first and last 32M
            with open(filename, 'rb') as f:
                while True:
                    data = f.read(block_size)
                    count -= 1
                    if count == 8:
                        # 2**23 = 8M
                        f.seek(-2**23, 2)
                    if not data or count == 0:
                        break
                    md5.update(data)
    except IOError as e:
        sys.exit('Failed to read {}: {}'.format(filename, e))
    return md5.hexdigest()


class BaseTarget:
    '''A base class for all targets (e.g. a file)'''
    def __init__(self):
        self._sigfile = None

    def exists(self, mode='any'):
        # mode should be 'any', 'target', or 'signature'
        raise RuntimeError('Undefined base function')

    def name(self):
        # name of the target, which should be able to differentiate
        # this object with other targets of the same type.
        raise RuntimeError('Undefined base function')

    def signature(self, mode='any'):
        # signature of the content of the target, which should be
        # able to detect changes of the content of target
        #
        # if mode == 'target', the target has to exist and the signature
        # has to be calculated. Otherwise you can return cached signature
        raise RuntimeError('Undefined base function')

    # -----------------------------------------------------
    # derived functions that do not need to be redefined
    #
    def sig_file(self):
        if self._sigfile is None:
            self._sigfile = os.path.join(env.exec_dir, '.sos', '.runtime', '{}_{}.file_info'.format(self.__class__.__name__,
                textMD5(self.name())))
        return self._sigfile

    def remove_sig(self):
        if self.sig_file() and os.path.isfile(self.sig_file()):
            os.remove(self.sig_file())

    def write_sig(self):
        '''Write .sig file with signature'''
        # path to file
        with open(self.sig_file(), 'w') as sig:
            sig.write('{}\t{}\n'.format(self.name(), self.signature()))

    def __repr__(self):
        return '{}("{}")'.format(self.__class__.__name__, self.name())

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, obj):
        return isinstance(obj, self.__class__) and self.signature() == obj.signature()

class sos_variable(BaseTarget):
    '''A target for a SoS variable.'''
    def __init__(self, var):
        super(sos_variable, self).__init__()
        self._var = var

    def exists(self, mode='any'):
        return self._var in env.sos_dict

    def name(self):
        return self._var

    def signature(self, mode='any'):
        return textMD5(self._var)

class env_variable(BaseTarget):
    '''A target for an environmental variable.'''
    def __init__(self, var):
        super(env_variable, self).__init__()
        self._var = var

    def exists(self, mode='any'):
        return self._var in os.environ

    def name(self):
        return self._var

    def signature(self, mode='any'):
        return textMD5(repr(os.environ[self._var]))

class sos_step(BaseTarget):
    '''A target for a step of sos.'''
    def __init__(self, step_name):
        super(sos_step, self).__init__()
        self._step_name = step_name

    def exists(self, mode='any'):
        # the target exists only if it has been executed?
        # which is indicated by a variable
        return '__completed__' in env.sos_dict and self._step_name in env.sos_dict['__completed__']

    def name(self):
        return self._step_name

    def signature(self, mode='any'):
        return textMD5('sos_step({})'.format(self._step_name))

    def write_sig(self):
        pass

# class bundle(BaseTarget):
#     '''a bundle of other targets'''
#     def __init__(self, *args):
#         super(bundle, self).__init__()
#         self._targets = []
#         for arg in args:
#             if isinstance(arg, str):
#                 self._targets.append(os.path.expanduser(arg))
#             elif isinstance(arg, Iterable):
#                 # in case arg is a Generator, check its type will exhaust it
#                 arg = list(arg)
#                 if not all(isinstance(x, str) for x in arg):
#                     raise RuntimeError('Invalid filename: {}'.format(arg))
#                 self._targets.extend(arg)
#             else:
#                 raise RuntimeError('Unrecognized file: {}'.format(arg))
#
#     def exists(self, mode='any'):
#         return all(FileTarget(x).exists(mode) for x in self._targets)
#
#     def name(self):
#         return repr(self._targets)
#
#     def signature(self, mode='any'):
#         return textMD5(repr(self._targets))
#
class dynamic(BaseTarget):
    '''A dynamic executable that only handles input files when
    it is available. This target is handled directly with its `resolve`
    function called by the executor. '''
    def __init__(self, target):
        self._target = target

    def name(self):
        return self._target

    def resolve(self):
        return self._target

class remote(BaseTarget):
    '''A remote target is not tracked and not translated during task execution'''
    def __init__(self, *targets):
        super(remote, self).__init__()
        self.__unresolvable_object__ = True
        if len(targets) == 1:
            self._target = targets[0]
        else:
            # multi-item targets
            self._target = targets
        if isinstance(self._target, Sequence) and not isinstance(self._target, str):
            self.__flattenable__ = True

    def name(self):
        if isinstance(self._target, str):
            return FileTarget(self._target).name()
        elif isinstance(self._target, BaseTarget):
            return self._target.name()
        else:
            return repr(self._target)

    def exists(self, mode='any'):
        return True

    def signature(self, mode='any'):
        return textMD5(self.name())

    def resolve(self):
        return self._target

    def flatten(self):
        return [remote(x) for x in self._target]

class executable(BaseTarget):
    '''A target for an executable command.'''

    def __init__(self, cmd, version=None):
        super(executable, self).__init__()
        self._cmd = cmd
        self._md5 = None
        if version is None:
            self._version = ()
        elif isinstance(version, str):
            self._version = (version,)
        else:
            self._version = tuple(version)

    def exists(self, mode='any'):
        if mode in ('any', 'target') and shutil.which(shlex.split(self._cmd)[0]):
            if self._version:
                import subprocess
                try:
                    output = subprocess.check_output(self._cmd,
                        stderr=subprocess.STDOUT, shell=True, timeout=5).decode()
                except subprocess.TimeoutExpired as e:
                    env.logger.warning(e)
                    return False
                except subprocess.CalledProcessError as e:
                    env.logger.warning(e)
                    return False
                for ver in self._version:
                    if ver in output:
                        return True
                return False
            else:
                return True
        if mode in ('any', 'signature') and os.path.isfile(self.sig_file()):
            return True
        return False

    def name(self):
        if self._version:
            return '{} (version={})'.format(self._cmd, self._version)
        else:
            return self._cmd

    def signature(self, mode='any'):
        if mode != 'target' and self._md5:
            return self._md5
        exe_file = shutil.which(shlex.split(self._cmd)[0])
        if exe_file is None or not os.path.isfile(exe_file):
            self._md5 = None
        else:
            self._md5 = fileMD5(exe_file)
        return self._md5

class FileTarget(BaseTarget):
    '''A regular target for files.
    '''
    def __init__(self, filename):
        super(FileTarget, self).__init__()
        self._filename = os.path.expanduser(filename)
        self._md5 = None
        self._attachments = []

    def exists(self, mode='any'):
        if mode in ('any', 'target') and os.path.isfile(self.fullname()):
            return True
        elif mode == 'any' and os.path.isfile(self.fullname() + '.zapped'):
            return True
        elif mode == 'signature' and os.path.isfile(self.sig_file()):
            return True
        return False

    def name(self):
        return self._filename

    # redefine sig_file because of special request to store sig files
    # in different folders
    def __repr__(self):
        return self.name()

    def sig_file(self):
        if self._sigfile is not None:
            return self._sigfile
        # If the output path is outside of the current working directory
        fullname = os.path.abspath(self.name())
        name_md5 = textMD5(fullname)

        if self.is_external():
            self._sigfile = os.path.join(os.path.expanduser('~'), '.sos', '.runtime', name_md5 + '.file_info')
        else:
            self._sigfile = os.path.join(env.exec_dir, '.sos', '.runtime', name_md5 + '.file_info')
        return self._sigfile

    def signature(self, mode='any'):
        '''Return file signature'''
        if mode == 'target':
            self._md5 = fileMD5(self.fullname())
        if self._md5 is not None:
            return self._md5
        if os.path.isfile(self.sig_file()) and (not os.path.isfile(self.fullname()) or os.path.getmtime(self.sig_file()) > os.path.getmtime(self.fullname())):
            with open(self.sig_file()) as md5:
                try:
                    line = md5.readline()
                    _, _, _, m = line.rsplit('\t', 3)
                    return m.strip()
                except Exception:
                    pass
        elif os.path.isfile(self.fullname() + '.zapped'):
            with open(self.fullname() + '.zapped') as md5:
                try:
                    line = md5.readline()
                    _, _, _, m = line.rsplit('\t', 3)
                    return m.strip()
                except Exception:
                    pass
        self._md5 = fileMD5(self.fullname())
        return self._md5
    #
    # FileTarget - specific functions. Not required by other targets
    #
    def add(self, filename):
        '''add related files to the same signature'''
        self._attachments.append(os.path.abspath(os.path.expanduser(filename)))

    def remove(self, mode='both'):
        if mode in ('both', 'target') and os.path.isfile(self.fullname()):
            os.remove(self.fullname())
        if mode in ('both', 'signature') and os.path.isfile(self.sig_file()):
            os.remove(self.sig_file())

    def is_external(self):
        try:
            return os.path.relpath(self.fullname(), env.exec_dir).startswith('..')
        except Exception:
            # under windows the file might be on different volume
            return True

    def fullname(self):
        return os.path.abspath(self.name())

    def size(self):
        if os.path.isfile(self._filename):
            return os.path.getsize(self.fullname())
        elif os.path.isfile(self.sig_file()):
            with open(self.sig_file()) as md5:
                line = md5.readline()
                _, _, s, _ = line.rsplit('\t', 3)
                return int(s.strip())
        elif os.path.isfile(self.fullname() + '.zapped'):
            with open(self.fullname() + '.zapped') as md5:
                line = md5.readline()
                _, _, s, _ = line.rsplit('\t', 3)
                return int(s.strip())
        else:
            raise RuntimeError('{} or its signature does not exist.'.format(self._filename))

    def mtime(self):
        if os.path.isfile(self._filename):
            return os.path.getmtime(self.fullname())
        elif os.path.isfile(self.sig_file()):
            with open(self.sig_file()) as md5:
                line = md5.readline()
                _, t, _, _ = line.rsplit('\t', 3)
                return t.strip()
        elif os.path.isfile(self.fullname() + '.zapped'):
            with open(self.fullname() + '.zapped') as md5:
                line = md5.readline()
                _, t, _, _ = line.rsplit('\t', 3)
                return t.strip()
        else:
            raise RuntimeError('{} or its signature does not exist.'.format(self._filename))

    def __eq__(self, other):
        return os.path.abspath(self.fullname()) == os.path.abspath(other.fullname())

    def write_sig(self):
        '''Write .file_info file with signature'''
        # path to file
        with open(self.sig_file(), 'w') as md5:
            md5.write('{}\t{}\t{}\t{}\n'.format(self.fullname(), os.path.getmtime(self.fullname()),
                os.path.getsize(self.fullname()), self.signature()))
            for f in self._attachments:
                md5.write('{}\t{}\t{}\t{}\n'.format(f, os.path.getmtime(f),
                    os.path.getsize(f), fileMD5(f)))

    def validate(self):
        '''Check if file matches its signature'''
        if not os.path.isfile(self.sig_file()):
            return False
        with open(self.sig_file()) as md5:
            for line in md5:
                f, _, _, m = line.rsplit('\t', 3)
                if not os.path.isfile(f):
                    return False
                if fileMD5(f) != m.strip():
                    env.logger.debug('MD5 mismatch {}'.format(f))
                    return False
        return True

class RuntimeInfo:
    '''Record run time information related to a number of output files. Right now only the
    .exe_info files are used.
    '''
    def __init__(self, step_md5, script, input_files=None, output_files=None, dependent_files = None,
        signature_vars = None, sdict=None):
        '''Runtime information for specified output files

        output_files:
            intended output file

        '''
        if sdict is None:
            sdict = env.sos_dict
        self.step_md5 = step_md5
        self.script = script
        # input can only be a list of files
        if not isinstance(input_files, list):
            if input_files is None:
                self.input_files = []
            else:
                raise RuntimeError('Input files must be a list of filenames for runtime signature.')
        else:
            self.input_files = [FileTarget(x) if isinstance(x, str) else x for x in input_files]

        if dependent_files is None:
            self.dependent_files = []
        elif isinstance(dependent_files, list):
            self.dependent_files = [FileTarget(x) if isinstance(x, str) else x for x in dependent_files]
        elif isinstance(dependent_files, Undetermined):
            self.dependent_files = dependent_files
        else:
            raise RuntimeError('Dependent files must be a list of filenames or Undetermined for runtime signature.')

        external_output = False
        if output_files is None:
            self.output_files = []
        elif isinstance(output_files, list):
            self.output_files = [FileTarget(x) if isinstance(x, str) else x for x in output_files]
            external_output = self.output_files and isinstance(self.output_files[0], FileTarget) and self.output_files[0].is_external()
        elif isinstance(output_files, Undetermined):
            self.output_files = output_files
        else:
            raise RuntimeError('Output files must be a list of filenames or Undetermined for runtime signature.')

        self.signature_vars = {} if signature_vars is None else {x: sdict[x] if x in sdict else Undetermined() for x in signature_vars}

        sig_vars = [] if signature_vars is None else sorted([x for x in signature_vars if x in sdict and isPrimitive(sdict[x])])
        self.sig_id = textMD5('{} {} {} {} {}'.format(self.script, self.input_files, output_files, self.dependent_files,
            '\n'.join('{}:{}'.format(x, stable_repr(sdict[x])) for x in sig_vars)))

        if external_output:
            # global signature
            self.proc_info = os.path.join(os.path.expanduser('~'), '.sos', '.runtime', '{}.exe_info'.format(self.sig_id))
        else:
            self.proc_info = os.path.join(env.exec_dir, '.sos', '.runtime', '{}.exe_info'.format(self.sig_id))

    def __getstate__(self):
        return {'step_md5': self.step_md5,
                'input_files': self.input_files,
                'output_files': self.output_files,
                'dependent_files': self.dependent_files,
                'signature_vars': self.signature_vars,
                'script': self.script}

    def __setstate__(self, sdict):
        self.step_md5 = sdict['step_md5']
        self.input_files = sdict['input_files']
        self.output_files = sdict['output_files']
        self.dependent_files = sdict['dependent_files']
        self.signature_vars = sdict['signature_vars']
        self.script = sdict['script']
        #
        # the signature might be on a remote machine and has changed location
        self.proc_info = os.path.join(os.path.expanduser('~'), '.sos', '.runtime', '{}.exe_info'.format(
            textMD5('{} {} {} {}'.format(self.script, self.input_files, self.output_files, self.dependent_files))))


    def lock(self):
        # we will need to lock on a file that we do not really write to
        # otherwise the lock will be broken when we write to it.
        self._lock = fasteners.InterProcessLock(self.proc_info + '_')
        if not self._lock.acquire(blocking=False):
            self._lock = None
            raise UnavailableLock((self.input_files, self.output_files, self.proc_info))
        else:
            env.logger.trace('Lock acquired for output files {}'.format(short_repr(self.output_files)))

    def release(self):
        if self._lock:
            try:
                self._lock.release()
                env.logger.trace('Lock released for output files {}'.format(short_repr(self.output_files)))
            except Exception as e:
                env.logger.warning('Unable to release lock for output files {}: {}'.format(self.output_files, e))
            finally:
                self._lock = None

    def set(self, files, file_type):
        # add signature file if input and output files are dynamic
        env.logger.trace('Set {} of signature to {}'.format(file_type, files))
        if file_type == 'output':
            self.output_files = [FileTarget(x) for x in files]
        elif file_type == 'depends':
            self.depends_files = [FileTarget(x) for x in files]
        else:
            raise RuntimeError('Invalid signature file type {}'.format(file_type))

    def write(self, rebuild=False):
        '''Write signature file with signature of script, input, output and dependent files.
        Because local input and output files can only be determined after the execution
        of workflow. They are not part of the construction...
        '''
        if isinstance(self.output_files, Undetermined) or isinstance(self.dependent_files, Undetermined):
            env.logger.trace('Write signature failed due to undetermined files')
            return False
        env.logger.trace('Write signature {}'.format(self.proc_info))
        with open(self.proc_info, 'w') as md5:
            md5.write('{}\n'.format(textMD5(self.script)))
            md5.write('# input\n')
            for f in self.input_files:
                if f.exists('target'):
                    # this calculates file MD5
                    f.write_sig()
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                elif not rebuild and f.exists('signature'):
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                else:
                    env.logger.warning('Failed to create signature: input target {} does not exist'.format(f))
                    return False
            md5.write('# output\n')
            for f in self.output_files:
                if f.exists('target'):
                    # this calculates file MD5
                    f.write_sig()
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                elif not rebuild and f.exists('signature'):
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                else:
                    env.logger.warning('Failed to create signature: output target {} does not exist'.format(f))
                    return False
            md5.write('# dependent\n')
            for f in self.dependent_files:
                if f.exists('target'):
                    # this calculates file MD5
                    f.write_sig()
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                elif not rebuild and f.exists('signature'):
                    md5.write('{}\t{}\n'.format(f, f.signature()))
                else:
                    env.logger.warning('Failed to create signature: dependent target {} does not exist'.format(f))
                    return False
            # context that will be needed for validation
            md5.write('# init context\n')
            for var in sorted(self.signature_vars.keys()):
                # var can be local and not passed as outside environment
                value = self.signature_vars[var]
                if not isinstance(value, Undetermined):
                    try:
                        md5.write(save_var(var, value))
                    except Exception:
                        env.logger.debug('Variable {} of value {} is ignored from step signature'.format(var, short_repr(value)))
            # context used to return context
            md5.write('# end context\n')
            for var in sorted(self.signature_vars.keys()):
                # var can be local and not passed as outside environment
                if var in env.sos_dict:
                    value = env.sos_dict[var]
                    try:
                        md5.write(save_var(var, value))
                    except Exception:
                        env.logger.debug('Variable {} of value {} is ignored from step signature'.format(var, short_repr(value)))
            md5.write('# step process\n')
            md5.write(self.script)
            md5.write('# step process\n')
            md5.write(self.script)
        # successfully write signature, write in workflow runtime info
        if '__workflow_sig__' in env.sos_dict and os.path.isfile(env.sos_dict['__workflow_sig__']):
            workflow_sig = env.sos_dict['__workflow_sig__']
            with fasteners.InterProcessLock(workflow_sig + '_'):
                with open(workflow_sig, 'a') as wf:
                    wf.write('EXE_SIG\tstep={}\tsession={}\n'.format(self.step_md5, os.path.basename(self.proc_info).split('.')[0]))
                    for f in self.input_files:
                        if isinstance(f, FileTarget):
                            wf.write('IN_FILE\tfilename={}\tsession={}\tsize={}\tmd5={}\n'.format(f, self.step_md5, f.size(), f.signature()))
                    for f in self.dependent_files:
                        if isinstance(f, FileTarget):
                            wf.write('IN_FILE\tfilename={}\tsession={}\tsize={}\tmd5={}\n'.format(f, self.step_md5, f.size(), f.signature()))
                    for f in self.output_files:
                        if isinstance(f, FileTarget):
                            wf.write('OUT_FILE\tfilename={}\tsession={}\tsize={}\tmd5={}\n'.format(f, self.step_md5, f.size(), f.signature()))
        return True

    def validate(self):
        '''Check if ofiles and ifiles match signatures recorded in md5file'''
        if not self.proc_info or not os.path.isfile(self.proc_info):
            return 'Missing signature file {}'.format(self.proc_info)
        env.logger.trace('Validating {}'.format(self.proc_info))
        #
        # file not exist?
        if isinstance(self.output_files, Undetermined):
            return "Undetermined output files"
        sig_files = self.input_files + self.output_files + self.dependent_files
        for x in sig_files:
            if not x.exists('any'):
                return 'Missing target {}'.format(x)
        #
        files_checked = {x.name():False for x in sig_files if not isinstance(x, Undetermined)}
        res = {'input': [], 'output': [], 'depends': [], 'vars': {}}
        cur_type = 'input'
        with open(self.proc_info) as md5:
            cmdMD5 = md5.readline().strip()   # command
            if textMD5(self.script) != cmdMD5:
                return "Changed command"
            for line in md5:
                if not line.strip():
                    continue
                if line.startswith('#'):
                    if line == '# input\n':
                        cur_type = 'input'
                    elif line == '# output\n':
                        cur_type = 'output'
                    elif line == '# dependent\n':
                        cur_type = 'depends'
                    elif line == '# init context\n':
                        cur_type = 'init context'
                    elif line == '# end context\n':
                        cur_type = 'end context'
                    elif line == '# step process\n':
                        break
                    else:
                        env.logger.trace('Unrecognized line in sig file {}'.format(line))
                    continue
                # for validation
                if cur_type == 'init context':
                    key, value = load_var(line)
                    if key not in env.sos_dict:
                        return 'Variable {} not in running environment'.format(key)
                    try:
                        try:
                            if env.sos_dict[key] != value:
                                return 'Context variable {} value mismatch: {} saved, {} current'.format(
                                    key, short_repr(value), short_repr(env.sos_dict[key]))
                        except Exception as e:
                            env.logger.warning("Variable {} of type {} cannot be compared: {}".format(key, type(key).__name__, e))
                    except Exception as e:
                        env.logger.warning('Failed to restore variable {} from signature: {}'.format(key, e))
                    continue
                # for return context
                elif cur_type == 'end context':
                    try:
                        key, value = load_var(line)
                        res['vars'][key] = value
                    except Exception as e:
                        env.logger.warning('Failed to restore variable {} from signature: {}'.format(key, e))
                    continue
                try:
                    f, m = line.rsplit('\t', 1)
                    if '(' in f and ')' in f:
                        # this part is hard, because this can be a customized target.
                        target_type = f.split('(')[0]
                        target_class = None
                        if target_type in globals():
                            target_class = eval(target_type)
                        else:
                            # check registry
                            for entrypoint in pkg_resources.iter_entry_points(group='sos_targets'):
                                if entrypoint.name.strip() == target_type:
                                    target_class = entrypoint.load()
                                    break
                        if target_class is None:
                            raise ValueError('Failed to identify target class {}'.format(target_type))
                        # parameter of class?
                        freal = eval(f, {target_type: target_class})
                    else:
                        freal = FileTarget(f)
                    if freal.exists('target'):
                        fmd5 = freal.signature('target')
                    elif freal.exists('signature'):
                        fmd5 = freal.signature()
                    else:
                        return 'File {} not exist'.format(f)
                    res[cur_type].append(freal.name() if isinstance(freal, FileTarget) else freal)
                    if fmd5 != m.strip():
                        return 'File has changed {}'.format(f)
                    files_checked[freal.name()] = True
                except Exception as e:
                    env.logger.debug('Wrong md5 line {} in {}: {}'.format(line, self.proc_info, e))
                    continue
        #
        if not all(files_checked.values()):
            return 'No MD5 signature for {}'.format(', '.join(x for x,y in files_checked.items() if not y))
        env.logger.trace('Signature matches and returns {}'.format(res))
        # validation success, record signature used
        if '__workflow_sig__' in env.sos_dict and os.path.isfile(env.sos_dict['__workflow_sig__']):
            workflow_sig = env.sos_dict['__workflow_sig__']
            with fasteners.InterProcessLock(workflow_sig + '_'):
                with open(workflow_sig, 'a') as wf:
                    wf.write(self.proc_info + '\n')
        return res

